/******************
Patrick Brodeur w/ WPI IQP team:
	¤ Reneé Lanza		¤ Lisa Morris		¤ Eddie Osowski
Single-file txt-to-csv parser
Purpose-built to read SWOMS data from email dumps into a CSV.
Written during November/December 2011
******************/

/********************
    Include Files
********************/
#include <windows.h>	// Windows includes; supersedes the others.
#include <stdio.h>
#include <string.h>
#include <time.h>

/****************
	Constants
****************/
#define L_DATETIME 11
#define D_NEWLINE 1
#define D_SPACE  2
#define D_M 3
#define CTS 1
#define NOT_CTS 0
#define BIGTABLE "SWOMS_DATA" // Change this to set the name of the table for syncing with DB

/**************************
	Function Prototypes
**************************/
fpos_t getData( char* outputStr, const char* sLabel, unsigned int maxLength, FILE* searchFile, unsigned int nDelimiter, unsigned char cts);
void getData2( char* outputStr, const char* sLabel, FILE* searchFile, char delimiter);
void textToFileName( char* outputStr );

/********************
	MAIN Function
********************/
int main(int argc, char *argv[])	// Setup for command-line input
{
	// Make sure there's an argument or else Windows (or  *nix) will cry. 
	if(argv[1] == NULL)
	{
		printf("Critical error: Must have input file on command line.\n");
		return 0;
	}
	
	/** Test for valid input **/
	if(fopen(argv[1], "r") == NULL)
	{
		printf("Must provide a valid input file\n");
		return 0;
	}
	
	/**********************
		Local Variables
	**********************/
	FILE *inFile, *outFile, *bigOutFile; // Pointers to the input file and output files
	FILE *SWOMSLOCK, *VTABLELOCK; // Pointers to the .LOCK files
	
	char tmpStr[100]; // Buffer string
	char outPath[175], outLock[180]; // Output directory path
	char imoNumber[] = "0000000";
	char* strPtr; // General pointer for string ops
	int i,j; // Iterators
	char date[L_DATETIME];
	char time[L_DATETIME];
	clock_t timeOut;
	
	/*****************
		Processing
	*****************/

	/** Open the input file **/
	strcpy(outPath, argv[1]); // Move into local variables
//printf("DEBUG: Input File is %s\n",outPath);
	inFile = fopen(outPath, "r"); // Open input file for reading
//printf("DEBUG: inPath is %s\n", outPath);
	
	
	/** Find working directory and establish output file **/
	getData(tmpStr, "VESSEL: ", 20, inFile, D_NEWLINE, 0); 	// Retrieve vessel's name from the input file
	textToFileName(tmpStr); // Remove prohibited characters
	strPtr = strrchr(outPath,'\\'); // Extract the path

	if(strPtr == NULL) strcpy(outPath, ".\\"); // If no path given, use working directory
	else outPath[strlen(outPath)-strlen(strPtr)+1] = 0; // Keep trailing slash
	strcat(outPath, tmpStr); // Add vessel name as the filename
	strcpy(outLock, outPath);
	strcat(outPath, ".csv"); // Tack on CSV extension so excel, etc. will recognize it
//printf("DEBUG: outPath is %s\n", outPath);
	
	/** Write first line if the file doesn't exist **/
	if( fopen(outPath, "r") == NULL)
	{
		outFile = fopen(outPath, "w");	// Create and open output file for writing
		getData(tmpStr, "VESSEL: ", 20, inFile, D_NEWLINE, 0); // Retrieve vessel name again
		fprintf(outFile, "0000000,%s\n", tmpStr); // Put the first line in (IMO Number, Shipname)
		fclose(outFile);
	}
	
/** Lock system: Lock down the vessel table **/
	j=0;
	printf("LOCKING: \n%s\n", outLock);
	strcat(outLock, ".LOCK");
	VTABLELOCK = fopen(outLock,"r");
	while(VTABLELOCK != NULL)
	{
		fclose(VTABLELOCK);
		j++;
		if(j>90)
		{
			printf("\nSWOMSPARSER FAILED. FILE:\n%s\n",argv[1]); 
			bigOutFile = fopen("ErrorLog.txt", "a");
			fprintf(bigOutFile,"Failed to lock Vessel table. Lock file: %s.\n", outLock);
			return 0;
		}
		if(j>=60) remove(outLock);
		printf("Failed to lock (%d). Retrying", j);
		if(j==30) printf("Warning: Attempting Lock Removal on Retry #60\n");
		timeOut = clock() + CLOCKS_PER_SEC;
		i=0;
		while (clock() < timeOut) 
		{
			if (clock() >= timeOut - ((10-i) * CLOCKS_PER_SEC / 10)) 
			{
				putchar('.');
				i++;
			}
		}
		putchar('\n');
		VTABLELOCK = fopen(outLock,"r");
	}

	VTABLELOCK = fopen(outLock,"w");
	fputs("LOCKED FOR EDITING",VTABLELOCK);
	fclose(VTABLELOCK);
	printf("LOCKED: \n%s\n\n", outLock);

	
	/** Retrieve the IMO number from line 1 **/
	outFile = fopen(outPath, "r"); 								// Open for reading
	fgets(tmpStr, 100, outFile); 								// Read in first line
	
	// Copy up until the first comma
	i=0; imoNumber[0]=tmpStr[0];
	while (imoNumber[i] != ',')
	{
		i++;
		imoNumber[i] = tmpStr[i];
	}
	imoNumber[i] = 0;
//printf("DEBUG: IMO NUMBER is %s\n", imoNumber);
	fclose(outFile);

	
	/** Lock system: Lock down the SWOMSTABLE table **/
	j=0;
	printf("LOCKING: \n%s\n", BIGTABLE".LOCK");
	SWOMSLOCK = fopen(BIGTABLE".LOCK","r");
	while(SWOMSLOCK != NULL)
	{
		fclose(SWOMSLOCK);
		j++;
		if(j>90)
		{
			printf("\nSWOMSPARSER FAILED. FILE:\n%s\n",argv[1]); 
			bigOutFile = fopen("ErrorLog.txt", "a");
			fprintf(bigOutFile,"Failed to lock SWOMS table. IMO: %s.\n", imoNumber);
			return 0;
		}
		if(j>=60) remove(BIGTABLE".csv");
		printf("Failed to lock (%d). Retrying\n", j);
		if(j==30) printf("Warning: Attempting Lock Removal on Retry #60\n");
		timeOut = clock() + CLOCKS_PER_SEC;
		i=0;
		while (clock() < timeOut) 
		{
			if (clock() == timeOut - ((10-i) * CLOCKS_PER_SEC / 10)) 
			{
				putchar('.');
				i++;
			}
		}
		putchar('\n');
		SWOMSLOCK = fopen(BIGTABLE".LOCK","r");
	}
	SWOMSLOCK = fopen(BIGTABLE".LOCK","w");
	fputs("LOCKED FOR EDITING",SWOMSLOCK);
	fclose(SWOMSLOCK);
	printf("LOCKED: \n%s\n\n", BIGTABLE".LOCK");

	
	/** Prepare individual and cumulative tables for appending data **/	
//printf("DEBUG: Output File -- %s\n", outPath);
	outFile = fopen(outPath, "a");
	if(strcmp(imoNumber,"0000000") != 0) bigOutFile = fopen(BIGTABLE".csv", "a+");
	
	
	/** Add IMO Number **/
	fprintf(outFile, "%s,", imoNumber);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,",imoNumber);
	
	
	/** Add Date and Time**/
	// Date... the hard part.
	date[0]=0;
	while ( date[0] != tmpStr[3] )
	{	
		// See if we've hit EOF yet...
		if( fgets(tmpStr, 100, inFile) == NULL )
		{
			printf("Error: Could not find date\n");
			return 0; 											// Returns function call, breaking the while loop and skipping the rest.
		}
		strPtr = strstr(tmpStr, "/");							// Find a forward slash.

		// See if the line goes like this: /##/####
		if( strPtr != NULL && strPtr[0] == '/' && 
			((strPtr[1] >= 48) && (strPtr[1] <= 57)) && ((strPtr[2] >= 48) && (strPtr[2] <= 57)) && (strPtr[3] == '/') && 
			((strPtr[4] >= 48) && (strPtr[4] <= 57)) && ((strPtr[5] >= 48) && (strPtr[5] <= 57)) && 
			((strPtr[6] >= 48) && (strPtr[6] <= 57)) && ((strPtr[7] >= 48) && (strPtr[7] <= 57)) )
		{
		// If it looks like a date, format it in MM/DD/YYYY
			date[0]=tmpStr[3]; date[1]=tmpStr[4]; date[3]=tmpStr[0]; date[4]=tmpStr[1];
			date[6]=tmpStr[6]; date[7]=tmpStr[7]; date[8]=tmpStr[8]; date[9]=tmpStr[9];
			date[2]='/'; date[5]='/'; 
			date[10]=0; 										// nul-terminate date string
//printf("DEBUG: Date is %s\n",date);
		}
	}
	// Write date to files
	fprintf(outFile, "%s,",date);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,",date);

	// Time, the part that comes right after the date.
	for(i=0; i<8; i++)
	{
		time[i]=tmpStr[11+i];
	}
	time[8] = 0;
//printf("DEBUG: Time is %s\n", time);
	
	// Write time to files
	fprintf(outFile, "%s,",time);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,",time);


	
	/** Get longitude, latitude, and all 24-hour data **/
	i=0;
	while(1)
	{
		// Acquire data from input file
		switch(i)
		{
		case  0: getData(tmpStr, "Latitude: ", 30, inFile, D_NEWLINE, 0); break;
		case  1: getData(tmpStr, "Longitude: ", 30, inFile, D_NEWLINE, 0); break;
		case  2: getData(tmpStr, "Blg Pmp Run ", 30, inFile, D_SPACE, 0); break;
		case  3: getData(tmpStr, "Blg Pmp Run ", 30, inFile, D_SPACE, 1); break;
		case  4: getData(tmpStr, "OWS Run ", 30, inFile, D_SPACE, 0); break;
		case  5: getData(tmpStr, "OWS Run ", 30, inFile, D_SPACE, 1); break;
		case  6: getData(tmpStr, "Overboard Open ", 30, inFile, D_SPACE, 0); break;
		case  7: getData(tmpStr, "Overboard Open ", 30, inFile, D_SPACE, 1); break;
		case  8: getData(tmpStr, "PPM Alarm ", 30, inFile, D_SPACE, 0); break;
		case  9: getData(tmpStr, "Average PPM Overboard ", 30, inFile, D_SPACE, 0); break;
		case 10: getData(tmpStr, "OCM Fresh Water Valve Opened ", 30, inFile, D_SPACE, 0); break;
		case 11: getData(tmpStr, "Oil Purge Valve Open ", 30, inFile, D_SPACE, 0); break;
		case 12: getData(tmpStr, "Oil Purge Valve Open ", 30, inFile, D_SPACE, 1); break;
		case 13: getData(tmpStr, "Incinerator Run ", 30, inFile, D_SPACE, 0); break;
		case 14: getData(tmpStr, "Incinerator Burning W.O. ", 30, inFile, D_SPACE, 0); break;
		default: goto tank_data;
		}
		
		// Put the data into the output files
		fprintf(outFile, "%s,", tmpStr);
		if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,", tmpStr);	// Only add to bigOutFile if the IMO number is non-zeros
//printf("DEBUG: Case: %d\t Value: %s\n", i, tmpStr);
		i++;
	}
	
	tank_data:
	/** Get current tank data **/
	// Phase 1: Read in first tank
	fseek(inFile,0L,SEEK_SET);
	getData2(tmpStr,"Lvl ",inFile,'m');
	fprintf(outFile, "%s,", tmpStr);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,", tmpStr);
	getData2(tmpStr,"Vol ",inFile,'m');
	fprintf(outFile, "%s,", tmpStr);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,", tmpStr);
	
	// Phase 2: Loop through tanks 2-15
	for(i=2; i<=15; i++)
	{
		getData2(tmpStr, "Lvl ", inFile, 'm');
		fprintf(outFile, "%s,", tmpStr);
		if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,", tmpStr);
		getData2(tmpStr, "Vol ", inFile, 'm');
		fprintf(outFile, "%s,", tmpStr);
		if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,", tmpStr);
	}
	
	// Phase 3: Read in tank 16 and prepare for next record
	getData2(tmpStr, "Lvl ", inFile, 'm');
	fprintf(outFile, "%s,", tmpStr);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s,", tmpStr);
	getData2(tmpStr, "Vol ", inFile, 'm');
	fprintf(outFile, "%s\n", tmpStr);
	if(strcmp(imoNumber,"0000000") != 0) fprintf(bigOutFile, "%s\n", tmpStr);

	fclose(outFile);
	if(strcmp(imoNumber,"0000000") != 0) fclose(bigOutFile);
	
	
	/** Lock system: Remove locks. **/
	if(strcmp(imoNumber,"0000000") != 0)
	{
		fclose(SWOMSLOCK);
		remove(BIGTABLE".LOCK");
	}
	fclose(VTABLELOCK);
	remove(outLock);

	return 0; // Successful exit.
		
}  // End main

/***************************
	Additional Functions
***************************/
// getData: Extracts bounded data from the text file being parsed, with a return value pointing just above where it finds data
fpos_t getData( char* outputStr, const char* sLabel, unsigned int maxLength, FILE* searchFile, unsigned int nDelimiter, unsigned char cts)
{
	// Local Variables
	unsigned int i;
	char tmpStr[100];	// Avoid segfaults; allocate the memory
	char *strPtr;
	fpos_t foundposition;
	
	// Default to "0" in case the program fails to find the search
	outputStr[0]='0';
	outputStr[1]=0;
	
	// Processing
	// Reset file position to beginning
	if(!cts) fseek(searchFile, 0L, SEEK_SET);

	while ( 1 )
	{	
		// Check if we hit EOF
		fgetpos(searchFile, &foundposition);
		if( fgets(tmpStr, 100, searchFile) == NULL )
			{
//printf("Error: Could not find text \"%s\"\n", sLabel);
				return foundposition; // Returns function call, breaking the while loop and skipping the rest.
//printf("Last Line Read: %s\n", tmpStr);
//printf("Output: %s\n", outputStr);
			}
		
		// Point to the section of tmpStr that has the data we care about
		strPtr = strstr(tmpStr, sLabel);
		
		// If found, stop searching.
		if (strPtr != NULL) break;	// Stop searching and continue on
	}

	// Update output string to contain the relevant data, up until the chosen delimiter
	for( i = 0; i < maxLength; i++ ) outputStr[i] = strPtr[i+strlen(sLabel)]; 
	switch (nDelimiter)
	{
	case D_NEWLINE:
		{
			i=0; while((outputStr[i] != '\n') && (outputStr[i] !=  '\r'))
			{
				i++; // Error if the i gets too high
				if(i==maxLength){ printf("Error: Bad Delimiter or Max Length\n\tNewline. Search Term: ---%s---\n", sLabel); return foundposition;}
			}
			outputStr[i]=0; break;
		}
	case D_SPACE:
		{
			i=0; 
			while((outputStr[i] != ' ') && (outputStr[i] != '\n') && (outputStr[i] !=  '\r'))
			{
				i++; // Error if the i gets too high
				if(i==maxLength){ printf("Error: Bad Delimiter or Max Length\nSpace. Search Term: ---%s---\n", sLabel); return foundposition;}
			}
			outputStr[i]=0; break;
		}
	case D_M:
		{
			i=0; 
			while((outputStr[i] != 'm') && (outputStr[i] != '\n') && (outputStr[i] !=  '\r'))
			{
				i++; // Error if the i gets too high
				if(i==maxLength){ printf("Error: Bad Delimiter or Max Length\n'm'. Search Term: ---%s---\n", sLabel); return foundposition;}
			}
			outputStr[i]=0; break;
		}
	default:printf("Error: Bad Delimiter Value\n\tSearch Term: ---%s---\n", sLabel); 
	}
//printf("Last Line Read: %s\n", tmpStr);
//printf("Output: %s\n", outputStr);
	return foundposition;
}

// textToFileName: Replaces any illegal characters with characters that are filename-friendly
void textToFileName( char* outputStr )
{
	unsigned int i; // Iterator
	//for( i=0; i<strlen(outputStr); i++ ) outputStr[i] = outputStr[i];
	for ( i=0; i < strlen(outputStr); i++ )
		{
		switch (outputStr[i])
			{
				// Remove all invalid characters
				case '\\': case '/': case '?': case '*': case '\"': case '<': case '>': case '|':
					outputStr[i] = '-'; break;
				case ' ': 
					outputStr[i] = '_'; break;
				default: break;
			}
		}
	outputStr[i] = 0; // End string with nul character

	return;
} // End textToFilename

// getData: Extracts bounded data from the text file being parsed, with a return value pointing just above where it finds data
void getData2( char* outputStr, const char* sLabel, FILE* searchFile, char delimiter)
{
	// Local Variables
	char tmpStr[100];	// Avoid segfaults; allocate the memory
	unsigned int i=0, j=0;
	
	// Default to "0" in case the program fails to find the search
	outputStr[0]='0';
	outputStr[1]=0;
	
	// Processing
	j = strlen(sLabel);
	while ( 1 )
	{	
		for (i=0; i <  j; i++) tmpStr[i] = tmpStr[i+1];
		tmpStr[j-1] = fgetc(searchFile);
		tmpStr[j] = 0;
		
//printf("%s", tmpStr);
//while(getchar() != '\n');
		if(strcmp(tmpStr, sLabel) == 0) break;
		if(tmpStr[j-1] == EOF)
		{
//printf("EOF Error\n");
			return;
		}
	}
	
	outputStr[0] = fgetc(searchFile); 
	i=0; while ( (outputStr[i] != delimiter) && (outputStr[i] != '\n') && (outputStr[i] !=  '\r') && (outputStr[i] !=  EOF))
	{	
		i++;
		outputStr[i] = fgetc(searchFile);
	}
	outputStr[i] = 0;
	
	i=0; while(outputStr[i] != 0)
	{
		switch(outputStr[i])
		{
		// as long as it's numeric or a decimal point, we don't care
		case '0': case '1': case '2': case '3': case '4': case '5':
		case '6': case '7': case '8': case '9': case '.': /*printf("%c\n",outputStr[i]); */break;
		
		// if it's not a number or decimal point, output "0"
		default: outputStr[0] = '0'; outputStr[1] = 0; /*printf("Set to Zero\n");*/ return;
		}
		i++;
	}

	return;
}	

	// End Program Source. 